#!/usr/bin/python3

import sys
import numpy as np
import numpysane as nps
import os
testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import mrcal._poseutils_npsp as _poseutils

import cv2
from testutils import *
from test_calibration_helpers import grad


def R_from_r(r):
    r'''Rotation matrix from a Rodrigues vector

    Simple reference implementation from wikipedia:

    https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
'''

    th_sq = nps.norm2(r)
    Kth = np.array(((    0, -r[2],  r[1]),
                     ( r[2],    0, -r[0]),
                     (-r[1], r[0],     0)))
    if th_sq > 1e-20:
        # normal path
        th = np.sqrt(th_sq)
        s = np.sin(th)
        c = np.cos(th)
        K = Kth  / th
        return np.eye(3) + s*K + (1. - c)*nps.matmult(K,K)

    # small th. Can't divide by it. But I can look at the limit.
    #
    # s*K = Kth * sin(th)/th -> Kth
    #
    # (1-c)*nps.matmult(K,K) = (1-c) Kth^2/th^2 -> Kth^2 s / 2th -> Kth^2/2
    return np.eye(3) + Kth + nps.matmult(Kth,Kth) / 2.


def r_from_R(R):
    r'''Rodrigues vector from a Rotation matrix

    Simple reference implementation from wikipedia:

    https://en.wikipedia.org/wiki/Rotation_matrix#Conversion_from_and_to_axis%E2%80%93angle

    I assume the input is a valid rotation matrix

    '''

    axis = np.array((R[2,1] - R[1,2],
                     R[0,2] - R[2,0],
                     R[1,0] - R[0,1] ))

    if nps.norm2(axis) > 1e-20:
        # normal path
        costh = (np.trace(R) - 1.)/2.
        th = np.arccos(costh)
        return axis / nps.mag(axis) * th

    # small mag(axis). Can't divide by it. But I can look at the limit.
    #
    # axis / (2 sinth)*th = axis/2 *th/sinth ~ axis/2
    return axis/2.


def Rt_from_rt(rt):
    r'''Simple reference implementation'''
    return nps.glue(R_from_r(rt[:3]), rt[3:], axis=-2)

def rt_from_Rt(Rt):
    r'''Simple reference implementation'''
    return nps.glue(r_from_R(Rt[:3,:]), Rt[3,:], axis=-1)


def invert_rt(rt):
    r'''Simple reference implementation

    b = Ra + t  -> a = R'b - R't
'''
    r = rt[:3]
    t = rt[3:]
    R = R_from_r(r)
    tinv = -nps.matmult(t, R)
    return nps.glue( -r, tinv.ravel(), axis=-1)

def invert_Rt(Rt):
    r'''Simple reference implementation

    b = Ra + t  -> a = R'b - R't
'''
    R = Rt[:3,:]
    tinv = -nps.matmult(Rt[3,:], R)
    return nps.glue( nps.transpose(R), tinv.ravel(), axis=-2)

def invert_R(R):
    r'''Simple reference implementation

'''
    return nps.transpose(R)

def compose_Rt(Rt0, Rt1):
    r'''Simple reference implementation

    b = R0 (R1 x + t1) + t0 =
      = R0 R1 x + R0 t1 + t0
'''
    R0 = Rt0[:3,:]
    t0 = Rt0[ 3,:]
    R1 = Rt1[:3,:]
    t1 = Rt1[ 3,:]
    R2 = nps.matmult(R0,R1)
    t2 = nps.matmult(t1, nps.transpose(R0)) + t0
    return nps.glue( R2, t2.ravel(), axis=-2)

def normalize_r(r):
    r'''If abs(rot) > 180deg -> flip direction, abs(rot) <- 180-abs(rot)'''
    norm2 = nps.norm2(r)
    if norm2 < np.pi*np.pi:
        return r

    mag = np.sqrt(norm2)
    r_unit = r / mag

    mag = mag % (np.pi*2.)
    if mag < np.pi:
        # same direction, but fewer full rotations
        return r_unit*mag

    return -r_unit * (np.pi*2. - mag)

def normalize_rt(rt):
    return nps.glue(normalize_r(rt[:3]),
                    rt[3:],
                    axis=-1)

def compose_r(r0, r1):
    r'''Simple reference implementation'''
    return r_from_R(nps.matmult( R_from_r(r0),
                                 R_from_r(r1)))

def compose_rt(rt0, rt1):
    r'''Simple reference implementation'''
    return rt_from_Rt(compose_Rt( Rt_from_rt(rt0),
                                  Rt_from_rt(rt1)))



# Big array. I'm going to slice this thing for my working arrays to produce
# interesting non-contiguous input, output
base   = np.zeros((7,11,13,5), dtype=float)



base[1,1,0:3,1] = np.array((1., 2., 0.1))
r0_ref = base[1,1,0:3,1]

base[1,1,3:6,1] = np.array((3., 5., -2.4))
t0_ref = base[1,1,3:6,1]

base[1,1,6:9,1] = np.array((-.3, -.2, 1.1))
r1_ref = base[1,1,6:9,1]

base[1,1,9:12,1] = np.array((-8.,  .5, -.4))
t1_ref = base[1,1,9:12,1]

base[1,2,0:3,1] = np.array((-10., -108., 3.))
x = base[1,2,0:3,1]

base[1,:3,:3,2] = R_from_r(r0_ref)
R0_ref = base[1,:3,:3,2]

base[1,3:7,:3,2]= nps.glue(R0_ref, t0_ref, axis=-2)
Rt0_ref = base[1,3:7,:3,2]

base[1,2,3:9,1]= nps.glue(r0_ref, t0_ref, axis=-1)
rt0_ref = base[1,2,3:9,1]

base[1,7:10,:3,2] = R_from_r(r1_ref)
R1_ref = base[1,7:10,:3,2]

base[2,:4,:3,2]= nps.glue(R1_ref, t1_ref, axis=-2)
Rt1_ref = base[2,:4,:3,2]

base[1,3,:6,1]= nps.glue(r1_ref, t1_ref, axis=-1)
rt1_ref = base[1,3,:6,1]


# the implementation has a separate path for tiny R, so I test it separately
base[1,5,0:3,1] = np.array((-2.e-18, 3.e-19, -5.e-18))
r0_ref_tiny = base[1,5,0:3,1]

base[5,:3,:3,2] = R_from_r(r0_ref_tiny)
R0_ref_tiny = base[5,:3,:3,2]



out333 = base[:3, :3,   :3,   3]
out343 = base[:3,3:7,   :3,   3]
out43  = base[2,  4:8,  :3,   2]
out33  = base[2,  8:11, :3,   2]
out33a = base[3,    :3, :3,   2]
out33b = base[3,   3:6, :3,   2]
out33c = base[3,   6:9, :3,   2]
out33d = base[4,   :3,  :3,   2]
out36  = base[6,   :3,  :6,   2]
out3   = base[1,  3,    6:9,  1]
out6   = base[1,  4,    :6,   1]
out66  = base[5,3:9,    3:9,  2]
out66a = base[6,3:9,    3:9,  2]

confirm_equal( mrcal.identity_R(out=out33),
               np.eye(3),
               msg='identity_R')
confirm_equal( mrcal.identity_Rt(out=out43),
               nps.glue(np.eye(3), np.zeros((3,),), axis=-2),
               msg='identity_Rt')
confirm_equal( mrcal.identity_r(out=out3),
               np.zeros((3,)),
               msg='identity_r')
confirm_equal( mrcal.identity_rt(out=out6),
               np.zeros((6,)),
               msg='identity_rt')

################# rotate_point_R
y = \
    mrcal.rotate_point_R(R0_ref, x, out = out3)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref)),
               msg='rotate_point_R result')

y, J_R, J_x = \
    mrcal.rotate_point_R(R0_ref, x, get_gradients=True,
                         out = (out3,out333,out33))
J_R_ref = grad(lambda R: nps.matmult(x, nps.transpose(R)),
               R0_ref)
J_x_ref = R0_ref
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref)),
               msg='rotate_point_R result')
confirm_equal( J_R,
               J_R_ref,
               msg='rotate_point_R J_R')
confirm_equal( J_x,
               J_x_ref,
               msg='rotate_point_R J_x')

# In-place
R0_ref_copy = np.array(R0_ref)
x_copy      = np.array(x)
y = \
    mrcal.rotate_point_R(R0_ref_copy, x_copy, out = x_copy)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref)),
               msg='rotate_point_R result written in-place into x')

# inverted
y = \
    mrcal.rotate_point_R(R0_ref, x, out = out3, inverted=True)
confirm_equal( y,
               nps.matmult(x, R0_ref),
               msg='rotate_point_R(inverted) result')

y, J_R, J_x = \
    mrcal.rotate_point_R(R0_ref, x, get_gradients=True,
                         out = (out3,out333,out33),
                         inverted = True)
J_R_ref = grad(lambda R: nps.matmult(x, R),
               R0_ref)
J_x_ref = nps.transpose(R0_ref)
confirm_equal( y,
               nps.matmult(x, R0_ref),
               msg='rotate_point_R(inverted) result')
confirm_equal( J_R,
               J_R_ref,
               msg='rotate_point_R(inverted) J_R')
confirm_equal( J_x,
               J_x_ref,
               msg='rotate_point_R(inverted) J_x')

# inverted in-place
R0_ref_copy = np.array(R0_ref)
x_copy      = np.array(x)
y = \
    mrcal.rotate_point_R(R0_ref_copy, x_copy, out = x_copy, inverted = True)
confirm_equal( y,
               nps.matmult(x, R0_ref),
               msg='rotate_point_R result written in-place into x')

################# rotate_point_r

y = mrcal.rotate_point_r(r0_ref, x, out = out3)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R_from_r(r0_ref))),
               msg='rotate_point_r result')

y, J_r, J_x = mrcal.rotate_point_r(r0_ref, x, get_gradients=True,
                                   out = (out3, out33, out33a))
J_r_ref = grad(lambda r: nps.matmult(x, nps.transpose(R_from_r(r))),
               r0_ref)
J_x_ref = grad(lambda x: nps.matmult(x, nps.transpose(R_from_r(r0_ref))),
               x)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R_from_r(r0_ref))),
               msg='rotate_point_r result')
confirm_equal( J_r,
               J_r_ref,
               msg='rotate_point_r J_r')
confirm_equal( J_x,
               J_x_ref,
               msg='rotate_point_r J_x')

r0_ref_copy  = np.array(r0_ref)
x_copy       = np.array(x)
y = mrcal.rotate_point_r(r0_ref_copy, x_copy,
                         out = x_copy)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R_from_r(r0_ref))),
               msg='rotate_point_r result written in-place into x')

r0_ref_copy  = np.array(r0_ref)
x_copy       = np.array(x)
out33_copy   = np.array(out33)
out33a_copy  = np.array(out33a)
y, J_r, J_x = mrcal.rotate_point_r(r0_ref_copy, x_copy, get_gradients=True,
                                   out = (x_copy, out33, out33a))
confirm_equal( y,
               nps.matmult(x, nps.transpose(R_from_r(r0_ref))),
               msg='rotate_point_r (with gradients) result written in-place into x')
confirm_equal( J_r,
               J_r_ref,
               msg='rotate_point_r (with gradients) result written in-place into x: J_r')
confirm_equal( J_x,
               J_x_ref,
               msg='rotate_point_r (with gradients) result written in-place into x: J_x')

# inverted
y = mrcal.rotate_point_r(r0_ref, x, out = out3, inverted=True)
confirm_equal( y,
               nps.matmult(x, R_from_r(r0_ref)),
               msg='rotate_point_r(inverted) result')

y, J_r, J_x = mrcal.rotate_point_r(r0_ref, x, get_gradients=True,
                                   out = (out3, out33, out33a),
                                   inverted = True)
J_r_ref = grad(lambda r: nps.matmult(x, R_from_r(r)),
               r0_ref)
J_x_ref = grad(lambda x: nps.matmult(x, R_from_r(r0_ref)),
               x)
confirm_equal( y,
               nps.matmult(x, R_from_r(r0_ref)),
               msg='rotate_point_r(inverted) result')
confirm_equal( J_r,
               J_r_ref,
               msg='rotate_point_r(inverted) J_r')
confirm_equal( J_x,
               J_x_ref,
               msg='rotate_point_r(inverted) J_x')

# inverted, in-place
r0_ref_copy  = np.array(r0_ref)
x_copy       = np.array(x)
y = mrcal.rotate_point_r(r0_ref_copy, x_copy, inverted = True,
                         out = x_copy)
confirm_equal( y,
               nps.matmult(x, R_from_r(r0_ref)),
               msg='rotate_point_r(inverted) result written in-place into x')
r0_ref_copy  = np.array(r0_ref)
x_copy       = np.array(x)
out33_copy   = np.array(out33)
out33a_copy  = np.array(out33a)
y, J_r, J_x = mrcal.rotate_point_r(r0_ref_copy, x_copy, get_gradients=True, inverted = True,
                                   out = (x_copy, out33, out33a))
confirm_equal( y,
               nps.matmult(x, R_from_r(r0_ref)),
               msg='rotate_point_r(inverted, with-gradients) result written in-place into x')
confirm_equal( J_r,
               J_r_ref,
               msg='rotate_point_r(inverted, with-gradients result written in-place into x) J_r')
confirm_equal( J_x,
               J_x_ref,
               msg='rotate_point_r(inverted, with-gradients result written in-place into x) J_x')


################# transform_point_Rt


y = mrcal.transform_point_Rt(Rt0_ref, x, out = out3)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_Rt result')

y, J_Rt, J_x = mrcal.transform_point_Rt(Rt0_ref, x, get_gradients=True,
                                        out = (out3, out343, out33))
J_Rt_ref = grad(lambda Rt: nps.matmult(x, nps.transpose(Rt[:3,:])) + Rt[3,:],
                Rt0_ref)
J_x_ref = R0_ref
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_Rt result')
confirm_equal( J_Rt,
               J_Rt_ref,
               msg='transform_point_Rt J_Rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_Rt J_x')

# In-place. I can imagine wanting to write the result in-place into t. But not
# into any of R
Rt0_ref_copy = np.array(Rt0_ref)
x_copy       = np.array(x)
y = mrcal.transform_point_Rt(Rt0_ref_copy, x_copy, out = Rt0_ref_copy[3,:])
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_Rt result written in-place into t')
Rt0_ref_copy = np.array(Rt0_ref)
x_copy       = np.array(x)
y = mrcal.transform_point_Rt(Rt0_ref_copy, x_copy, out = x_copy)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_Rt result written in-place into x')

Rt0_ref_copy = np.array(Rt0_ref)
x_copy       = np.array(x)
y, J_Rt, J_x = mrcal.transform_point_Rt(Rt0_ref_copy, x_copy,
                                        out = (Rt0_ref_copy[3,:], out343, out33),
                                        get_gradients = True)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_Rt (with gradients) result written in-place into t')
confirm_equal( J_Rt,
               J_Rt_ref,
               msg='transform_point_Rt (with gradients) result written in-place into t: J_Rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_Rt (with gradients) result written in-place into t: J_x')
Rt0_ref_copy = np.array(Rt0_ref)
x_copy       = np.array(x)
y, J_Rt, J_x = mrcal.transform_point_Rt(Rt0_ref_copy, x_copy,
                                        out = (x_copy, out343, out33),
                                        get_gradients = True)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_Rt (with gradients) result written in-place into x')
confirm_equal( J_Rt,
               J_Rt_ref,
               msg='transform_point_Rt (with gradients) result written in-place into x: J_Rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_Rt (with gradients) result written in-place into x: J_x')

# inverted
y = mrcal.transform_point_Rt(Rt0_ref, x, out = out3,
                             inverted = True)
confirm_equal( y,
               nps.matmult(x, R0_ref) - nps.matmult(t0_ref, R0_ref),
               msg='transform_point_Rt(inverted) result')

y, J_Rt, J_x = mrcal.transform_point_Rt(Rt0_ref, x, get_gradients=True,
                                        out = (out3, out343, out33),
                                        inverted = True)
J_Rt_ref = grad(lambda Rt: nps.matmult(x, Rt[:3,:]) - nps.matmult(Rt[3,:], Rt[:3,:]),
                Rt0_ref)
J_x_ref = nps.transpose(R0_ref)
confirm_equal( y,
               nps.matmult(x, R0_ref)-nps.matmult(t0_ref, R0_ref),
               msg='transform_point_Rt(inverted) result')
confirm_equal( J_Rt,
               J_Rt_ref,
               msg='transform_point_Rt(inverted) J_Rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_Rt(inverted) J_x')

# inverted in-place. I can imagine wanting to write the result in-place into t. But not
# into any of R
Rt0_ref_copy = np.array(Rt0_ref)
x_copy       = np.array(x)
y = mrcal.transform_point_Rt(Rt0_ref_copy, x_copy, out = Rt0_ref_copy[3,:],
                             inverted = True)
confirm_equal( y,
               nps.matmult(x, R0_ref)-nps.matmult(t0_ref, R0_ref),
               msg='transform_point_Rt(inverted) result written in-place into t')
Rt0_ref_copy = np.array(Rt0_ref)
x_copy       = np.array(x)
y = mrcal.transform_point_Rt(Rt0_ref_copy, x_copy, out = x_copy,
                             inverted = True)
confirm_equal( y,
               nps.matmult(x, R0_ref)-nps.matmult(t0_ref, R0_ref),
               msg='transform_point_Rt(inverted) result written in-place into x')

Rt0_ref_copy = np.array(Rt0_ref)
x_copy       = np.array(x)
y, J_Rt, J_x = mrcal.transform_point_Rt(Rt0_ref_copy, x_copy,
                                        out = (Rt0_ref_copy[3,:], out343, out33),
                                        get_gradients = True,
                                        inverted = True)
confirm_equal( y,
               nps.matmult(x, R0_ref)-nps.matmult(t0_ref, R0_ref),
               msg='transform_point_Rt(inverted) (with gradients) result written in-place into t')
confirm_equal( J_Rt,
               J_Rt_ref,
               msg='transform_point_Rt(inverted) (with gradients) result written in-place into t: J_Rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_Rt(inverted) (with gradients) result written in-place into t: J_x')
Rt0_ref_copy = np.array(Rt0_ref)
x_copy       = np.array(x)
y, J_Rt, J_x = mrcal.transform_point_Rt(Rt0_ref_copy, x_copy,
                                        out = (x_copy, out343, out33),
                                        get_gradients = True,
                                        inverted = True)
confirm_equal( y,
               nps.matmult(x, R0_ref)-nps.matmult(t0_ref, R0_ref),
               msg='transform_point_Rt(inverted) (with gradients) result written in-place into x')
confirm_equal( J_Rt,
               J_Rt_ref,
               msg='transform_point_Rt(inverted) (with gradients) result written in-place into x: J_Rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_Rt(inverted) (with gradients) result written in-place into x: J_x')


################# transform_point_rt


y = mrcal.transform_point_rt(rt0_ref, x, out = out3)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_rt result')

y, J_rt, J_x = mrcal.transform_point_rt(rt0_ref, x, get_gradients=True,
                                        out = (out3,out36,out33a))
J_rt_ref = grad(lambda rt: nps.matmult(x, nps.transpose(R_from_r(rt[:3])))+rt[3:],
                rt0_ref)
J_x_ref = grad(lambda x: nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               x)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_rt result')
confirm_equal( J_rt,
               J_rt_ref,
               msg='transform_point_rt J_rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_rt J_x')

# In-place. I can imagine wanting to write the result in-place into t. But not
# into any of r or J
rt0_ref_copy = np.array(rt0_ref)
x_copy       = np.array(x)
y = mrcal.transform_point_rt(rt0_ref_copy, x_copy,
                             out = rt0_ref_copy[3:])
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_rt result written in-place into t')
rt0_ref_copy = np.array(rt0_ref)
x_copy       = np.array(x)
y = mrcal.transform_point_rt(rt0_ref_copy, x_copy,
                             out = x_copy)
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_rt result written in-place into x')

rt0_ref_copy = np.array(rt0_ref)
x_copy       = np.array(x)
out36_copy   = np.array(out36)
out33a_copy  = np.array(out33a)
y, J_rt, J_x = mrcal.transform_point_rt(rt0_ref_copy, x_copy, get_gradients=True,
                                        out = (rt0_ref_copy[3:],out36_copy,out33a_copy))
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_rt (with gradients) result written in-place into t')
confirm_equal( J_rt,
               J_rt_ref,
               msg='transform_point_rt (with gradients) result written in-place into t: J_rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_rt (with gradients) result written in-place into t: J_x')
rt0_ref_copy = np.array(rt0_ref)
x_copy       = np.array(x)
out36_copy   = np.array(out36)
out33a_copy  = np.array(out33a)
y, J_rt, J_x = mrcal.transform_point_rt(rt0_ref_copy, x_copy, get_gradients=True,
                                        out = (x_copy, out36_copy,out33a_copy))
confirm_equal( y,
               nps.matmult(x, nps.transpose(R0_ref))+t0_ref,
               msg='transform_point_rt (with gradients) result written in-place into x')
confirm_equal( J_rt,
               J_rt_ref,
               msg='transform_point_rt (with gradients) result written in-place into x: J_rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_rt (with gradients) result written in-place into x: J_x')

# Inverted
y = mrcal.transform_point_rt(rt0_ref, x, out = out3, inverted=True)
confirm_equal( y,
               nps.matmult(x-t0_ref, R0_ref),
               msg='transform_point_rt(inverted) result')

y, J_rt, J_x = mrcal.transform_point_rt(rt0_ref, x, get_gradients=True,
                                        out = (out3,out36,out33a),
                                        inverted = True)
J_rt_ref = grad(lambda rt: nps.matmult(x-rt[3:], R_from_r(rt[:3])),
                rt0_ref)
J_x_ref = grad(lambda x: nps.matmult(x-t0_ref, R0_ref),
               x)
confirm_equal( y,
               nps.matmult(x-t0_ref, R0_ref),
               msg='transform_point_rt(inverted) result')
confirm_equal( J_rt,
               J_rt_ref,
               msg='transform_point_rt(inverted) J_rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_rt(inverted) J_x')

# Inverted in-place. I can imagine wanting to write the result in-place into t.
# But not into any of r or J
rt0_ref_copy = np.array(rt0_ref)
x_copy       = np.array(x)
y = mrcal.transform_point_rt(rt0_ref_copy, x_copy, inverted=True,
                             out = rt0_ref_copy[3:])
confirm_equal( y,
               nps.matmult(x-t0_ref, R0_ref),
               msg='transform_point_rt(inverted) result written in-place into t')
rt0_ref_copy = np.array(rt0_ref)
x_copy       = np.array(x)
y = mrcal.transform_point_rt(rt0_ref_copy, x_copy, inverted=True,
                             out = x_copy)
confirm_equal( y,
               nps.matmult(x-t0_ref, R0_ref),
               msg='transform_point_rt(inverted) result written in-place into x')

rt0_ref_copy = np.array(rt0_ref)
x_copy       = np.array(x)
out36_copy   = np.array(out36)
out33a_copy  = np.array(out33a)
y, J_rt, J_x = mrcal.transform_point_rt(rt0_ref_copy, x_copy, get_gradients=True, inverted=True,
                                        out = (rt0_ref_copy[3:],out36_copy,out33a_copy))
confirm_equal( y,
               nps.matmult(x-t0_ref, R0_ref),
               msg='transform_point_rt(inverted, with gradients) result written in-place into t')
confirm_equal( J_rt,
               J_rt_ref,
               msg='transform_point_rt(inverted, with gradients) result written in-place into t: J_rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_rt(inverted, with gradients) result written in-place into t: J_x')
rt0_ref_copy = np.array(rt0_ref)
x_copy       = np.array(x)
out36_copy   = np.array(out36)
out33a_copy  = np.array(out33a)
y, J_rt, J_x = mrcal.transform_point_rt(rt0_ref_copy, x_copy, get_gradients=True, inverted=True,
                                        out = (x_copy,out36_copy,out33a_copy))
confirm_equal( y,
               nps.matmult(x-t0_ref, R0_ref),
               msg='transform_point_rt(inverted, with gradients) result written in-place into x')
confirm_equal( J_rt,
               J_rt_ref,
               msg='transform_point_rt(inverted, with gradients) result written in-place into x: J_rt')
confirm_equal( J_x,
               J_x_ref,
               msg='transform_point_rt(inverted, with gradients) result written in-place into x: J_x')


################# r_from_R


r = mrcal.r_from_R(R0_ref, out = out3)
confirm_equal( r,
               r0_ref,
               msg='r_from_R result')

r, J_R = mrcal.r_from_R(R0_ref, get_gradients=True,
                        out = (out3,out333))
J_R_ref = grad(r_from_R,
               R0_ref)
confirm_equal( r,
               r0_ref,
               msg='r_from_R result')
confirm_equal( J_R,
               J_R_ref,
               msg='r_from_R J_R')

# Do it again, actually calling opencv. This is both a test, and shows how to
# migrate old code
r, J_R = mrcal.r_from_R(R0_ref, get_gradients=True,
                        out = (out3,out333))
rref,J_R_ref = cv2.Rodrigues(R0_ref)
confirm_equal( r,
               rref,
               msg='r_from_R result, comparing with cv2.Rodrigues')

# I'm not comparing with opencv's gradient report or dr/dR. It doesn't match. I
# know my gradient is correct because I numerically checked it above. Maybe
# opencv is doing something different because of the constraints placed on R.
# Figuring this out would be good
# J_R_ref = nps.transpose(J_R_ref) # fix opencv's weirdness. Now shape=(3,9)
# J_R_ref = J_R_ref.reshape(3,3,3)
# confirm_equal( J_R,
#                J_R_ref,
#                msg='r_from_R J_R, comparing with cv2.Rodrigues')


# I've seen this show up in the wild. r_from_R() was producing [nan nan nan]
R_fuzzed_I = \
    np.array([[ 0.9999999999999999              , -0.000000000000000010408340855861,  0.                              ],
              [-0.000000000000000010408340855861,  0.9999999999999999              ,  0.000000000000000013877787807814],
              [ 0.                              ,  0.000000000000000013877787807814,  0.9999999999999999              ]])
confirm_equal( mrcal.r_from_R(R_fuzzed_I), np.zeros((3,)),
               msg = 'r_from_R() can handle numerical fuzz')

################# R_from_r
R = mrcal.R_from_r(r0_ref, out = out33)
confirm_equal( R,
               R0_ref,
               msg='R_from_r result')

R, J_r = mrcal.R_from_r(r0_ref, get_gradients=True,
                        out = (out33,out333))
J_r_ref = grad(R_from_r,
               r0_ref)
confirm_equal( R,
               R0_ref,
               msg='R_from_r result')
confirm_equal( J_r,
               J_r_ref,
               msg='R_from_r J_r')

# the implementation has a separate path for tiny R, so I test it separately
R = mrcal.R_from_r(r0_ref_tiny, out = out33)
confirm_equal( R,
               R0_ref_tiny,
               msg='R_from_r result for tiny r0')

R, J_r = mrcal.R_from_r(r0_ref_tiny, get_gradients=True,
                        out = (out33,out333))
J_r_ref = grad(R_from_r,
               r0_ref_tiny)
confirm_equal( R,
               R0_ref_tiny,
               msg='R_from_r result for tiny r0')
confirm_equal( J_r,
               J_r_ref,
               msg='R_from_r J_r for tiny r0')

# Do it again, actually calling opencv. This is both a test, and shows how to
# migrate old code
R, J_r = mrcal.R_from_r(r0_ref, get_gradients=True,
                        out = (out33,out333))
Rref,J_r_ref = cv2.Rodrigues(r0_ref)
J_r_ref = nps.transpose(J_r_ref) # fix opencv's weirdness. Now shape=(9,3)
J_r_ref = J_r_ref.reshape(3,3,3)
confirm_equal( R,
               Rref,
               msg='R_from_r result, comparing with cv2.Rodrigues')
confirm_equal( J_r,
               J_r_ref,
               msg='R_from_r J_r, comparing with cv2.Rodrigues')



rt = mrcal.rt_from_Rt(Rt0_ref, out = out6)
confirm_equal( rt,
               rt0_ref,
               msg='rt_from_Rt result')

rt, J_R = mrcal.rt_from_Rt(Rt0_ref, get_gradients = True,
                           out = (out6,out333))
J_R_ref = grad(r_from_R,
               Rt0_ref[:3,:])
confirm_equal( rt,
               rt0_ref,
               msg='rt_from_Rt result')
confirm_equal( J_R,
               J_R_ref,
               msg='rt_from_Rt grad result')

Rt = mrcal.Rt_from_rt(rt0_ref, out=out43)
confirm_equal( Rt,
               Rt0_ref,
               msg='Rt_from_rt result')

Rt, J_r = mrcal.Rt_from_rt(rt0_ref, get_gradients = True,
                           out = (out43,out333))
J_r_ref = grad(R_from_r,
               rt0_ref[:3])
confirm_equal( Rt,
               Rt0_ref,
               msg='Rt_from_rt result')
confirm_equal( J_r,
               J_r_ref,
               msg='Rt_from_rt grad result')

Rt = mrcal.invert_Rt(Rt0_ref, out=out43)
confirm_equal( Rt,
               invert_Rt(Rt0_ref),
               msg='invert_Rt result')

# in-place
Rt0_ref_copy = np.array(Rt0_ref)
Rt = mrcal.invert_Rt(Rt0_ref_copy, out=Rt0_ref_copy)
confirm_equal( Rt,
               invert_Rt(Rt0_ref),
               msg='invert_Rt result written in-place')

R = mrcal.invert_R(R0_ref, out=out33)
confirm_equal( R,
               invert_R(R0_ref),
               msg='invert_R result')

# in-place
R0_ref_copy = np.array(R0_ref)
R = mrcal.invert_R(R0_ref_copy, out=R0_ref_copy)
confirm_equal( R,
               invert_R(R0_ref),
               msg='invert_R result written in-place')

rt = mrcal.invert_rt(rt0_ref, out=out6)
confirm_equal( rt,
               invert_rt(rt0_ref),
               msg='invert_rt result')

# in-place
rt0_ref_copy = np.array(rt0_ref)
rt = mrcal.invert_rt(rt0_ref_copy, out=rt0_ref_copy)
confirm_equal( rt,
               invert_rt(rt0_ref),
               msg='invert_rt result written in-place')

rt,drt_drt  = mrcal.invert_rt(rt0_ref,
                              get_gradients = True,
                              out=(out6,out66))
drt_drt_ref = grad(invert_rt,
                   rt0_ref)
confirm_equal( rt,
               invert_rt(rt0_ref),
               msg='invert_rt with grad result')
confirm_equal( drt_drt,
               drt_drt_ref,
               msg='invert_rt drt/drt result')

# in-place
rt0_ref_copy = np.array(rt0_ref)
drt_drt_copy = np.array(drt_drt)
rt,drt_drt = mrcal.invert_rt(rt0_ref_copy, out=(rt0_ref_copy,drt_drt_copy),
                             get_gradients=True)
confirm_equal( rt,
               invert_rt(rt0_ref),
               msg='invert_rt with grad result written in-place')
confirm_equal( drt_drt,
               drt_drt_ref,
               msg='invert_rt with grad drt/drt result written in-place')

Rt2 = mrcal.compose_Rt(Rt0_ref, Rt1_ref,
                       out=out43)
confirm_equal( Rt2,
               compose_Rt(Rt0_ref, Rt1_ref),
               msg='compose_Rt result')

# in-place
Rt0_ref_copy = np.array(Rt0_ref)
Rt1_ref_copy = np.array(Rt1_ref)
Rt2 = mrcal.compose_Rt(Rt0_ref_copy, Rt1_ref_copy,
                       out=Rt0_ref_copy)
confirm_equal( Rt2,
               compose_Rt(Rt0_ref, Rt1_ref),
               msg='compose_Rt result written in-place to Rt0')
Rt0_ref_copy = np.array(Rt0_ref)
Rt1_ref_copy = np.array(Rt1_ref)
Rt2 = mrcal.compose_Rt(Rt0_ref_copy, Rt1_ref_copy,
                       out=Rt1_ref_copy)
confirm_equal( Rt2,
               compose_Rt(Rt0_ref, Rt1_ref),
               msg='compose_Rt result written in-place to Rt1')




rt2 = mrcal.compose_rt(rt0_ref, rt1_ref, out = out6)
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt result')

# in-place
rt0_ref_copy = np.array(rt0_ref)
rt1_ref_copy = np.array(rt1_ref)
rt2 = mrcal.compose_rt(rt0_ref_copy, rt1_ref_copy,
                       out=rt0_ref_copy)
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt result written in-place to rt0')
rt0_ref_copy = np.array(rt0_ref)
rt1_ref_copy = np.array(rt1_ref)
rt2 = mrcal.compose_rt(rt0_ref_copy, rt1_ref_copy,
                       out=rt1_ref_copy)
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt result written in-place to rt1')

rt2 = _poseutils._compose_rt(rt0_ref, rt1_ref, out=out6)
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt result; calling _compose_rt() directly')

# in-place
rt0_ref_copy = np.array(rt0_ref)
rt1_ref_copy = np.array(rt1_ref)
rt2 = _poseutils._compose_rt(rt0_ref_copy, rt1_ref_copy,
                             out=rt0_ref_copy)
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt (calling _compose_rt() directly) written in-place to rt0')
rt0_ref_copy = np.array(rt0_ref)
rt1_ref_copy = np.array(rt1_ref)
rt2 = _poseutils._compose_rt(rt0_ref_copy, rt1_ref_copy,
                             out=rt1_ref_copy)
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt (calling _compose_rt() directly) written in-place to rt1')

rt2,drt2_drt0,drt2_drt1 = \
    mrcal.compose_rt(rt0_ref, rt1_ref, get_gradients=True,
                     out = (out6, out66, out66a))
drt2_drt0_ref = grad(lambda rt0: compose_rt( rt0, rt1_ref), rt0_ref)
drt2_drt1_ref = grad(lambda rt1: compose_rt( rt0_ref, rt1), rt1_ref)
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt result')
confirm_equal( drt2_drt0,
               drt2_drt0_ref,
               msg='compose_rt drt2_drt0')
confirm_equal( drt2_drt1,
               drt2_drt1_ref,
               msg='compose_rt drt2_drt1')

# in-place
rt0_ref_copy = np.array(rt0_ref)
rt1_ref_copy = np.array(rt1_ref)
drt2_drt0_copy = np.array(drt2_drt0)
drt2_drt1_copy = np.array(drt2_drt1)
rt2,drt2_drt0,drt2_drt1 = \
    mrcal.compose_rt(rt0_ref_copy, rt1_ref_copy, get_gradients=True,
                     out=(rt0_ref_copy,drt2_drt0_copy,drt2_drt1_copy))
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt (with gradients) result written in-place to rt0')
confirm_equal( drt2_drt0,
               drt2_drt0_ref,
               msg='compose_rt (with gradients) result written in-place to rt0: drt2_drt0')
confirm_equal( drt2_drt1,
               drt2_drt1_ref,
               msg='compose_rt (with gradients) result written in-place to rt0: drt2_drt1')
rt0_ref_copy = np.array(rt0_ref)
rt1_ref_copy = np.array(rt1_ref)
rt2,drt2_drt0,drt2_drt1 = \
    mrcal.compose_rt(rt0_ref_copy, rt1_ref_copy, get_gradients=True,
                     out=(rt1_ref_copy,drt2_drt0_copy,drt2_drt1_copy))
confirm_equal( rt2,
               compose_rt(rt0_ref, rt1_ref),
               msg='compose_rt (with gradients) result written in-place to rt1')
confirm_equal( drt2_drt0,
               drt2_drt0_ref,
               msg='compose_rt (with gradients) result written in-place to rt1: drt2_drt0')
confirm_equal( drt2_drt1,
               drt2_drt1_ref,
               msg='compose_rt (with gradients) result written in-place to rt1: drt2_drt1')

Rt2 = mrcal.compose_Rt(Rt0_ref, Rt1_ref,Rt0_ref,
                       out=out43)
confirm_equal( Rt2,
               compose_Rt(compose_Rt(Rt0_ref, Rt1_ref), Rt0_ref),
               msg='compose_Rt with 3 inputs')

rt2 = mrcal.compose_rt(rt0_ref, rt1_ref,rt0_ref,
                       out=out6)
# Needed here. The two rotations are semantically equivalent, but numerically
# different
rt2 = normalize_rt(rt2)
confirm_equal( rt2,
               compose_rt(compose_rt(rt0_ref, rt1_ref), rt0_ref),
               msg='compose_rt with 3 inputs')

################# compose_r()
#
# I check big, almost-0 and 0 arrays in both positions, and all the gradients.
r0big      = base[:3,0,0,0]
r0nearzero = base[:3,1,0,0]
r0zero     = base[:3,2,0,0]
r1big      = base[:3,3,0,0]
r1nearzero = base[:3,4,0,0]
r1zero     = base[:3,5,0,0]
r01        = base[:3,6,0,0]
dr01_dr0   = base[3:6,:3,0,0]
dr01_dr1   = base[3:6,3:6,0,0]

r0big[:] = np.array(( 1.0,  3.0, 1.1)) * 1e-1
r1big[:] = np.array((-2.0, -1.2, 0.3)) * 1e-1

r0nearzero[:] = np.array(( 1.7, -2.0, -5.1 )) * 1e-12
r1nearzero[:] = np.array((-1.2,  5.2,  0.03)) * 1e-12

r0zero[:] *= 0.
r1zero[:] *= 0.

confirm_equal( mrcal.compose_r(r0big, r1big),
               compose_r      (r0big, r1big),
               msg='compose_r basic operation')

mrcal.compose_r(r0big, r1big, out=r01)
confirm_equal( r01,
               compose_r      (r0big, r1big),
               msg='compose_r in-place output')

r01_notinplace, dr01_dr0_notinplace, dr01_dr1_notinplace = \
    mrcal.compose_r(r0big, r1big, get_gradients=True)
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1big), r0big, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0big, r1), r1big, step=1e-5)
confirm_equal( r01_notinplace,
               compose_r      (r0big, r1big),
               msg='compose_r gradients: r01')
confirm_equal( dr01_dr0_notinplace,
               dr01_dr0_ref,
               msg='compose_r gradients: dr01_dr0')
confirm_equal( dr01_dr1_notinplace,
               dr01_dr1_ref,
               msg='compose_r gradients: dr01_dr1')

mrcal.compose_r(r0big, r1big,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
confirm_equal( r01,
               compose_r      (r0big, r1big),
               msg='compose_r in-place gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place gradients: dr01_dr1')


mrcal.compose_r(r0big, r1nearzero,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1nearzero), r0big, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0big, r1), r1nearzero, step=1e-5)
confirm_equal( r01,
               compose_r      (r0big, r1nearzero),
               msg='compose_r in-place r1nearzero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r1nearzero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r1nearzero gradients: dr01_dr1')

mrcal.compose_r(r0big, r1zero,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1zero), r0big, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0big, r1), r1zero, step=1e-5)
confirm_equal( r01,
               compose_r      (r0big, r1zero),
               msg='compose_r in-place r1zero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r1zero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r1zero gradients: dr01_dr1')

mrcal.compose_r(r0nearzero, r1big,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1big), r0nearzero, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0nearzero, r1), r1big, step=1e-5)
confirm_equal( r01,
               compose_r      (r0nearzero, r1big),
               msg='compose_r in-place r1nearzero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r1nearzero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r1nearzero gradients: dr01_dr1')

mrcal.compose_r(r0zero, r1big,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1big), r0zero, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0zero, r1), r1big, step=1e-5)
confirm_equal( r01,
               compose_r      (r0zero, r1big),
               msg='compose_r in-place r1zero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r1zero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r1zero gradients: dr01_dr1')

mrcal.compose_r(r0nearzero, r1nearzero,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1nearzero), r0nearzero, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0nearzero, r1), r1nearzero, step=1e-5)
confirm_equal( r01,
               compose_r      (r0nearzero, r1nearzero),
               msg='compose_r in-place r0nearzero,r1nearzero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r0nearzero,r1nearzero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r0nearzero,r1nearzero gradients: dr01_dr1')

mrcal.compose_r(r0nearzero, r1zero,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1zero), r0nearzero, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0nearzero, r1), r1zero, step=1e-5)
confirm_equal( r01,
               compose_r      (r0nearzero, r1zero),
               msg='compose_r in-place r0nearzero,r1zero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r0nearzero,r1zero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r0nearzero,r1zero gradients: dr01_dr1')

mrcal.compose_r(r0zero, r1nearzero,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1nearzero), r0zero, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0zero, r1), r1nearzero, step=1e-5)
confirm_equal( r01,
               compose_r      (r0zero, r1nearzero),
               msg='compose_r in-place r0zero,r1nearzero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r0zero,r1nearzero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r0zero,r1nearzero gradients: dr01_dr1')

mrcal.compose_r(r0zero, r1zero,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1zero), r0zero, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0zero, r1), r1zero, step=1e-5)
confirm_equal( r01,
               compose_r      (r0zero, r1zero),
               msg='compose_r in-place r0zero,r1zero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r0zero,r1zero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r0zero,r1zero gradients: dr01_dr1')

# Finally, let's look at rotation composition when the result is 0
mrcal.compose_r(r0big, -r0big,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, -r0big), r0big, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0big, r1), -r0big, step=1e-5)
confirm_equal( r01,
               compose_r      (r0big, -r0big),
               msg='compose_r in-place r0big,-r0big gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r0big,-r0big gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r0big,-r0big gradients: dr01_dr1')

mrcal.compose_r(r0nearzero, -r0nearzero,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, -r0nearzero), r0nearzero, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0nearzero, r1), -r0nearzero, step=1e-5)
confirm_equal( r01,
               compose_r      (r0nearzero, -r0nearzero),
               msg='compose_r in-place r0nearzero,-r0nearzero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r0nearzero,-r0nearzero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r0nearzero,-r0nearzero gradients: dr01_dr1')

mrcal.compose_r(r0zero, -r0zero,
                get_gradients=True,
                out = (r01, dr01_dr0, dr01_dr1))
dr01_dr0_ref = grad(lambda r0: compose_r( r0, -r0zero), r0zero, step=1e-5)
dr01_dr1_ref = grad(lambda r1: compose_r( r0zero, r1), -r0zero, step=1e-5)
confirm_equal( r01,
               compose_r      (r0zero, -r0zero),
               msg='compose_r in-place r0zero,-r0zero gradients: r01')
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r in-place r0zero,-r0zero gradients: dr01_dr0')
confirm_equal( dr01_dr1,
               dr01_dr1_ref,
               msg='compose_r in-place r0zero,-r0zero gradients: dr01_dr1')

################# compose_r_tinyr0_gradientr0()
# These are a subset of the compose_r() tests just above
r1big      = base[:3,3,0,0]
r1nearzero = base[:3,4,0,0]
r0zero     = base[:3,2,0,0]
r1zero     = base[:3,5,0,0]
r01        = base[:3,6,0,0]
dr01_dr0   = base[3:6,:3,0,0]

r1big      [:] = np.array((-2.0, -1.2, 0.3)) * 1e-1
r1nearzero [:] = np.array((-1.2,  5.2,  0.03)) * 1e-12
r0zero     [:] *= 0.
r1zero     [:] *= 0.


mrcal.compose_r_tinyr0_gradientr0(r1big,
                                  out = dr01_dr0)
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1big), r0zero, step=1e-5)
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r_tinyr0_gradientr0 in-place r1zero gradients: dr01_dr0')

mrcal.compose_r_tinyr0_gradientr0(r1nearzero,
                                  out = dr01_dr0)
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1nearzero), r0zero, step=1e-5)
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r_tinyr0_gradientr0 in-place r1nearzero gradients: dr01_dr0')

mrcal.compose_r_tinyr0_gradientr0(r1zero,
                                  out = dr01_dr0)
dr01_dr0_ref = grad(lambda r0: compose_r( r0, r1zero), r0zero, step=1e-5)
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r_tinyr0_gradientr0 in-place r1zero gradients: dr01_dr0')

mrcal.compose_r_tinyr0_gradientr0(-r0zero,
                                  out = dr01_dr0)
dr01_dr0_ref = grad(lambda r0: compose_r( r0, -r0zero), r0zero, step=1e-5)
confirm_equal( dr01_dr0,
               dr01_dr0_ref,
               msg='compose_r_tinyr0_gradientr0 in-place r0zero,-r0zero gradients: dr01_dr0')

finish()



#!/usr/bin/python3

r'''Triangulation and projection uncertainty broadcasting test

The uncertainty routines support broadcasting, which we evaluate here

'''

import sys
import argparse
import re
import os

def parse_args():

    parser = \
        argparse.ArgumentParser(description = __doc__,
                                formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--fixed',
                        type=str,
                        choices=('cam0','frames'),
                        default = 'cam0',
                        help='''Are we putting the origin at camera0, or are all the frames at a fixed (and
                        non-optimizeable) pose? One or the other is required.''')
    parser.add_argument('--model',
                        type=str,
                        choices=('opencv4','opencv8','splined'),
                        default = 'opencv4',
                        help='''Which lens model we're using. Must be one of
                        ('opencv4','opencv8','splined')''')
    parser.add_argument('--Nframes',
                        type=int,
                        default=50,
                        help='''How many chessboard poses to simulate. These are dense observations: every
                        camera sees every corner of every chessboard pose''')
    parser.add_argument('--stabilize-coords',
                        action = 'store_true',
                        help='''Whether we report the triangulation in the camera-0 coordinate system (which
                        is moving due to noise) or in a stabilized coordinate
                        system based on the frame poses''')
    parser.add_argument('--cull-left-of-center',
                        action = 'store_true',
                        help='''If given, the calibration data in the left half of the imager is thrown
                        out''')
    parser.add_argument('--q-calibration-stdev',
                        type    = float,
                        default = 0.0,
                        help='''The observed pixel uncertainty of the chessboard
                        observations at calibration time. Defaults to 0.0. At
                        least one of --q-calibration-stdev and
                        --q-observation-stdev MUST be given as > 0''')
    parser.add_argument('--q-observation-stdev',
                        type    = float,
                        default = 0.0,
                        help='''The observed pixel uncertainty of the point
                        observations at triangulation time. Defaults to 0.0. At
                        least one of --q-calibration-stdev and
                        --q-observation-stdev MUST be given as > 0''')
    parser.add_argument('--q-observation-stdev-correlation',
                        type    = float,
                        default = 0.0,
                        help='''By default, the noise in the observation-time
                        pixel observations is assumed independent. This isn't
                        entirely realistic: observations of the same feature in
                        multiple cameras originate from an imager correlation
                        operation, so they will have some amount of correlation.
                        If given, this argument specifies how much correlation.
                        This is a value in [0,1] scaling the stdev. 0 means
                        "independent" (the default). 1.0 means "100%
                        correlated".''')
    parser.add_argument('--baseline',
                        type    = float,
                        default = 2.,
                        help='''The baseline of the camera pair. This is the
                        horizontal distance between each pair of adjacent
                        cameras''')

    args = parser.parse_args()

    if args.q_calibration_stdev <= 0.0 and \
       args.q_observation_stdev <= 0.0:
        raise Exception('At least one of --q-calibration-stdev and --q-observation-stdev MUST be given as > 0')
    return args


args = parse_args()

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils
import copy
import numpy as np
import numpysane as nps

from test_calibration_helpers import calibration_baseline



Ncameras = 3
Npoints  = 5

# shape (Npoints,3)
p_triangulated_true0 = np.zeros((Npoints,3), dtype=float)
p_triangulated_true0[:,0] = np.arange(Npoints)       - 2.77
p_triangulated_true0[:,1] = np.arange(Npoints) / 10. - 0.344
p_triangulated_true0[:,2] = 100.

############# Set up my world, and compute all the perfect positions, pixel
############# observations of everything
fixedframes = (args.fixed == 'frames')
object_spacing          = 0.1
object_width_n          = 10
object_height_n         = 9
calobject_warp_true     = np.array((0.002, -0.005))

# I want the RNG to be deterministic
np.random.seed(0)


extrinsics_rt_fromref_true = np.zeros((Ncameras,6), dtype=float)
extrinsics_rt_fromref_true[:,:3] = np.random.randn(Ncameras,3) * 0.1
extrinsics_rt_fromref_true[:, 3] = args.baseline * np.arange(Ncameras)
extrinsics_rt_fromref_true[:,4:] = np.random.randn(Ncameras,2) * 0.1

# cam0 is at the identity. This makes my life easy: I can assume that the
# optimization_inputs returned by calibration_baseline() use the same ref
# coordinate system as these transformations. I explicitly state this by passing
# calibration_baseline(allow_nonidentity_cam0_transform=False) later
extrinsics_rt_fromref_true[0] *= 0

optimization_inputs_baseline,                          \
models_true, models_baseline,                          \
indices_frame_camintrinsics_camextrinsics,             \
lensmodel, Nintrinsics, imagersizes,                   \
intrinsics_true, extrinsics_true_mounted, frames_true, \
observations_true,                                     \
args.Nframes =                                         \
    calibration_baseline(args.model,
                         Ncameras,
                         args.Nframes,
                         None,
                         object_width_n,
                         object_height_n,
                         object_spacing,
                         extrinsics_rt_fromref_true,
                         calobject_warp_true,
                         fixedframes,
                         testdir,
                         cull_left_of_center = args.cull_left_of_center,
                         allow_nonidentity_cam0_transform = False)



####################################################################
# I have the calibrated models, and I can compute the triangulations
#
# I triangulate my 5 points while observed by cameras (2,1) and cameras (1,0)
icameras    = ( (2,1), (1,0) )
Nmodelpairs = len(icameras)

# shape (Nmodelpairs,2)
M = [ [models_baseline[i] for i in icameras[0]],
      [models_baseline[i] for i in icameras[1]] ]

# shape (Npoints,Nmodelpairs, 2,2)
q = np.zeros((Npoints,Nmodelpairs, 2,2), dtype=float)
for ipt in range(Npoints):
    for imp in range(Nmodelpairs):
        q[ipt,imp,0,:] = mrcal.project(p_triangulated_true0[ipt],
                                       *M[imp][0].intrinsics())
        q[ipt,imp,1,:] = mrcal.project(mrcal.transform_point_Rt( mrcal.compose_Rt( M[imp][1].extrinsics_Rt_fromref(),
                                                                                   M[imp][0].extrinsics_Rt_toref() ),
                                                                 p_triangulated_true0[ipt]),
                                       *M[imp][1].intrinsics())

p, \
Var_p0p1_calibration_big, \
Var_p0p1_observation_big, \
Var_p0p1_joint_big = \
    mrcal.triangulate( q, M,
                       q_calibration_stdev             = args.q_calibration_stdev,
                       q_observation_stdev             = args.q_observation_stdev,
                       q_observation_stdev_correlation = args.q_observation_stdev_correlation,
                       stabilize_coords                = args.stabilize_coords )

testutils.confirm_equal(p.shape,
                        (Npoints,Nmodelpairs,3),
                        msg = "point array has the right shape")

testutils.confirm_equal(Var_p0p1_calibration_big.shape,
                        (Npoints,Nmodelpairs,3, Npoints,Nmodelpairs,3),
                        msg = "Big covariance (calibration) matrix has the right shape")
testutils.confirm_equal(Var_p0p1_observation_big.shape,
                        (Npoints,Nmodelpairs,3,3),
                        msg = "Big covariance (observation) matrix has the right shape")
testutils.confirm_equal(Var_p0p1_joint_big.shape,
                        (Npoints,Nmodelpairs,3, Npoints,Nmodelpairs,3),
                        msg = "Big covariance (joint) matrix has the right shape")

# Now I check each block in the diagonal individually
for ipt in range(Npoints):
    for imp in range(Nmodelpairs):
        p, \
        _, _, \
        Var_p0p1 = \
            mrcal.triangulate( q[ipt,imp], M[imp],
                               q_calibration_stdev             = args.q_calibration_stdev,
                               q_observation_stdev             = args.q_observation_stdev,
                               q_observation_stdev_correlation = args.q_observation_stdev_correlation,
                               stabilize_coords                = args.stabilize_coords )

        testutils.confirm_equal(Var_p0p1_joint_big[ipt,imp,:,ipt,imp,:],
                                Var_p0p1,
                                msg = f"Covariance (joint) sub-matrix ipt={ipt} imp={imp}")
        testutils.confirm_equal(Var_p0p1_calibration_big[ipt,imp,:,ipt,imp,:] + Var_p0p1_observation_big[ipt,imp,:,:],
                                Var_p0p1,
                                msg = f"Covariance (cal + obs) sub-matrix ipt={ipt} imp={imp}")

testutils.finish()



#!/usr/bin/python3

import sys
import numpy as np
import numpysane as nps
import os
import re
from inspect import currentframe

import mrcal.utils

Nchecks = 0
NchecksFailed = 0

# no line breaks. Useful for test reporting. Yes, this sets global state, but
# we're running a test harness. This is fine
np.set_printoptions(linewidth=1e10, suppress=True)

def test_location():
    r'''Reports string describing current location in the test'''


    filename_this = os.path.split( __file__ )[1]

    frame = currentframe().f_back.f_back

    # I keep popping the stack until I leave the testutils file and I'm not in a
    # function called "check"
    while frame:
        if frame.f_back is None or \
           (not frame.f_code.co_filename.endswith(filename_this) and
            frame.f_code.co_name != "check" ):
            break
        frame = frame.f_back

    testfile = os.path.split(frame.f_code.co_filename)[1]
    try:
        return "{}:{}".format(testfile, frame.f_lineno)
    except:
        return ''


def print_red(x):
    """print the message in red"""
    sys.stdout.write("\x1b[31m" + test_location() + ": " + x + "\x1b[0m\n")


def print_green(x):
    """Print the message in green"""
    sys.stdout.write("\x1b[32m" + test_location() + ": " + x + "\x1b[0m\n")

def print_blue(x):
    """Print the message in blue"""
    sys.stdout.write("\x1b[34m" + test_location() + ": " + x + "\x1b[0m\n")



def relative_scale(a,b, eps = 1e-6):
    return (np.abs(a) + np.abs(b)) / 2 + eps

def relative_diff(a,b, eps = 1e-6):
    return (a - b) / relative_scale(a,b, eps)

def confirm_equal(x, xref, msg='',
                  eps=1e-6,
                  reldiff_eps = 1e-6,
                  relative=False,
                  worstcase=False,
                  percentile=None):
    r'''If x is equal to xref, report test success.

    msg identifies this check. eps sets the RMS equality tolerance. The x,xref
    arguments can be given as many different types. This function tries to do
    the right thing.

    if relative: I look at a relative error:
                 err = (a-b) / ((abs(a)+abs(b))/2 + eps)
    else:        I look at absolute error:
                 err = a-b

    if worstcase: I look at the worst-case error
                  error = np.max(np.abs(err))
    elif percentile is not None: I look at the given point in the error distribution
                  error = np.percentile(np.abs(err), percentile)
    else:         RMS error
                  error = np.sqrt(nps.norm2(err) / len(err))
    '''

    global Nchecks
    global NchecksFailed
    Nchecks = Nchecks + 1

    # strip all trailing whitespace in each line, in case these are strings
    if isinstance(x, str):
        x = re.sub('[ \t]+(\n|$)', '\\1', x)
    if isinstance(xref, str):
        xref = re.sub('[ \t]+(\n|$)', '\\1', xref)

    # convert data to numpy if possible
    try:
        xref = np.array(xref)
    except:
        pass
    try:
        x = np.array(x)
    except:
        pass

    try:  # flatten array if possible
        x = x.ravel()
        xref = xref.ravel()
    except:
        pass

    try:
        N = x.shape[0]
    except:
        N = 1
    try:
        Nref = xref.shape[0]
    except:
        Nref = 1

    if N != Nref:

        # Comparing an array to a scalar reference is allowed
        if Nref == 1:
            xref = np.ones((N,), dtype=float) * xref
            Nref = N
        else:
            print_red(("FAILED{}: mismatched array sizes: N = {} but Nref = {}. Arrays: \n" +
                       "x = {}\n" +
                       "xref = {}").
                      format((': ' + msg) if msg else '',
                             N, Nref,
                             x, xref))
            NchecksFailed = NchecksFailed + 1
            return False

    if N != 0:
        try:  # I I can subtract, get the error that way
            if relative:
                diff = relative_diff(x, xref, eps = reldiff_eps)
            else:
                diff = x - xref

            if worstcase:
                what = 'worst-case'
                err  = np.max(np.abs(diff))
            elif percentile is not None:
                what = f'{percentile}%-percentile'
                err  = np.percentile(np.abs(diff), percentile, interpolation='higher')
            else:
                what = 'RMS'
                err  = np.sqrt(nps.norm2(diff) / len(diff))

            if not np.all(np.isfinite(err)):
                print_red(f"FAILED{(': ' + msg) if msg else ''}: Some comparison results are NaN or Inf. {what}. error_x_xref =\n{nps.cat(err,x,xref)}")
                NchecksFailed = NchecksFailed + 1
                return False
            if err > eps:
                print_red(f"FAILED{(': ' + msg) if msg else ''}: {what} error = {err}. x_xref_err =\n{nps.cat(x,xref,diff)}")
                NchecksFailed = NchecksFailed + 1
                return False
        except:  # Can't subtract. Do == instead
            if not np.array_equal(x, xref):
                print_red(f"FAILED{(': ' + msg) if msg else ''}: x_xref =\n{nps.cat(x,xref)}")
                NchecksFailed = NchecksFailed + 1
                return False
    print_green("OK" + (': ' + msg) if msg else '')
    return True


def confirm(x, msg=''):
    r'''If x is true, report test success.

    msg identifies this check'''

    global Nchecks
    global NchecksFailed
    Nchecks = Nchecks + 1

    if not x:
        print_red("FAILED{}".format((': ' + msg) if msg else ''))
        NchecksFailed = NchecksFailed + 1
        return False
    print_green("OK{}".format((': ' + msg) if msg else ''))
    return True


def confirm_raises(f, msg=''):
    r'''If f() raises an exception, report test success.

    msg identifies this check'''

    global Nchecks
    global NchecksFailed
    Nchecks = Nchecks + 1

    try:
        f()
        print_red("FAILED{}".format((': ' + msg) if msg else ''))
        NchecksFailed = NchecksFailed + 1
        return False
    except:
        print_green("OK{}".format((': ' + msg) if msg else ''))
        return True


def confirm_does_not_raise(f, msg=''):
    r'''If f() raises an exception, report test failure.

    msg identifies this check'''

    global Nchecks
    global NchecksFailed
    Nchecks = Nchecks + 1

    try:
        f()
        print_green("OK{}".format((': ' + msg) if msg else ''))
        return True
    except:
        print_red("FAILED{}".format((': ' + msg) if msg else ''))
        NchecksFailed = NchecksFailed + 1
        return False


def confirm_covariances_equal(var, var_ref,
                              what,
                              # scalar float to use for all the eigenvalues, of
                              # a list of length 3, to use in order from largest
                              # to smallest. None to skip that axis
                              eps_eigenvalues,
                              eps_eigenvectors_deg):

    # First, the thing is symmetric, right?
    confirm_equal(nps.transpose(var),
                  var,
                  worstcase = True,
                  msg = f"Var(dq) is symmetric for {what}")


    l_predicted,v_predicted = mrcal.utils._sorted_eig(var)
    l_observed,v_observed   = mrcal.utils._sorted_eig(var_ref)

    eccentricity_predicted = l_predicted[-1] / l_predicted[-2]

    for i in range(var.shape[-1]):
        # check all the eigenvalues, in order from largest to smallest
        if isinstance(eps_eigenvalues, float):
            eps = eps_eigenvalues
        else:
            eps = eps_eigenvalues[i]
            if eps is None:
                continue

        confirm_equal(l_observed[-1-i],
                      l_predicted[-1-i],
                      eps = eps,
                      worstcase = True,
                      relative  = True,
                      msg = f"Var(dq) worst[{i}] eigenvalue match for {what}")

    # I only check the eigenvector directions if the ellipse is sufficiently
    # non-circular. A circular ellipse has poorly-defined eigenvector directions
    if eccentricity_predicted > 2.:

        # I look at the direction of the largest ellipse axis only
        v0_predicted = v_predicted[:,-1]
        v0_observed  = v_observed [:,-1]

        confirm_equal(np.arcsin(nps.mag(np.cross(v0_observed,v0_predicted))) * 180./np.pi,
                      0,
                      eps = eps_eigenvectors_deg,
                      worstcase = True,
                      msg = f"Var(dq) eigenvectors match for {what}")

    # I don't bother checking v1. I already made sure the matrix is
    # symmetric. Thus the eigenvectors are orthogonal, so any angle offset
    # in v0 will be exactly the same in v1



def finish():
    r'''Finalize the executed tests.

    Prints the test summary. Exits successfully iff all the tests passed.

    '''
    if not Nchecks and not NchecksFailed:
        print_red("No tests defined")
        sys.exit(0)

    if NchecksFailed:
        print_red("Some tests failed: {} out of {}".format(NchecksFailed, Nchecks))
        sys.exit(1)

    print_green("All tests passed: {} total".format(Nchecks))
    sys.exit(0)



#!/usr/bin/python3

r'''Tests various pose transformations'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils


p = np.array([[1.79294696, 3.84665205, 8.88059626],
              [1.94285588, 5.5994178,  9.08980913],
              [1.16926135, 7.11203249, 6.49111756],
              [4.97016243, 3.36377894, 9.91309303],
              [7.0578383,  6.96290783, 2.74822211],
              [2.09286004, 8.32613405, 4.29534757],
              [5.390182,   6.97763688, 8.75498418],
              [6.54142328, 3.77331963, 3.29282044],
              [5.84571708, 4.87274621, 4.32639624],
              [7.00320821, 0.70546965, 5.67190652]])

Rt = np.array([[ 0.37796712, -0.5149426 ,  0.7693991 ],
               [ 0.52322138,  0.80441527,  0.28134581],
               [-0.76379333,  0.29622659,  0.5734715 ],
               [ 0.9447233,   6.8439095,   9.6958398 ]])
R = Rt[:3, :]
t = Rt[ 3, :]

noise = np.array([[0.00035356, 0.00043613, 0.00006606],
                  [0.00043968, 0.00043783, 0.00060678],
                  [0.00063803, 0.00024423, 0.00010871],
                  [0.00004966, 0.00053377, 0.00018905],
                  [0.00007708, 0.00023529, 0.0002229 ],
                  [0.00090558, 0.00072379, 0.00004062],
                  [0.00072059, 0.00074467, 0.00044128],
                  [0.00024228, 0.00058201, 0.00041458],
                  [0.00018121, 0.00078172, 0.00016128],
                  [0.00019021, 0.00001371, 0.00096808]])

Tp = nps.matmult( p, nps.transpose(R) ) + t
Rt_fit = \
    mrcal.align_procrustes_points_Rt01(Tp + noise,
                                       p)
R_fit = Rt_fit[:3, :]
t_fit = Rt_fit[ 3, :]
testutils.confirm_equal( R_fit, R, eps=1e-2, msg='Procrustes fit R' )
testutils.confirm_equal( t_fit, t, eps=1e-2, msg='Procrustes fit t' )

R_fit_vectors = \
    mrcal.align_procrustes_vectors_R01(nps.matmult( p, nps.transpose(R) ) + noise,
                                       p)
testutils.confirm_equal( R_fit_vectors, R, eps=1e-2, msg='Procrustes fit R (vectors)' )


testutils.confirm_equal( mrcal.invert_Rt(mrcal.Rt_from_rt(mrcal.invert_rt(mrcal.rt_from_Rt(Rt)))), Rt,
                         msg = 'Rt/rt and invert')

testutils.confirm_equal( mrcal.compose_Rt( Rt, mrcal.invert_Rt(Rt)),
                         nps.glue(np.eye(3), np.zeros((3,)), axis=-2),
                         msg = 'compose_Rt')

testutils.confirm_equal( mrcal.compose_rt( mrcal.rt_from_Rt(Rt), mrcal.invert_rt(mrcal.rt_from_Rt(Rt))),
                         np.zeros((6,)),
                         msg = 'compose_rt')

testutils.confirm_equal( mrcal.compose_r( mrcal.r_from_R(R),
                                          -mrcal.r_from_R(R)),
                         np.zeros((3,)),
                         msg = 'compose_r')

testutils.confirm_equal( mrcal.identity_Rt(),
                         nps.glue(np.eye(3), np.zeros((3,)), axis=-2),
                         msg = 'identity_Rt')

testutils.confirm_equal( mrcal.identity_rt(),
                         np.zeros((6,)),
                         msg = 'identity_rt')

testutils.confirm_equal( mrcal.transform_point_Rt( Rt, p ),
                         Tp,
                         msg = 'transform_point_Rt')

testutils.confirm_equal( mrcal.transform_point_rt( mrcal.rt_from_Rt(Rt), p ),
                         Tp,
                         msg = 'transform_point_rt')

testutils.confirm_equal( mrcal.transform_point_Rt( mrcal.invert_Rt(Rt), Tp ),
                         p,
                         msg = 'transform_point_Rt inverse')

testutils.confirm_equal( mrcal.transform_point_rt( mrcal.invert_rt(mrcal.rt_from_Rt(Rt)), Tp ),
                         p,
                         msg = 'transform_point_rt inverse')

######### quaternion business
testutils.confirm_equal( mrcal.R_from_quat( mrcal.quat_from_R(R) ),
                         R,
                         msg = 'R <-> quaternion transforms are inverses of one another')

# shape (2,3,3)
RR = nps.cat( R, nps.matmult(R,R) )
# shape (2,4,3)
RtRt = nps.cat( Rt, mrcal.compose_Rt(Rt,Rt) )
testutils.confirm_equal( mrcal.R_from_quat( mrcal.quat_from_R(RR) ),
                         RR,
                         msg = 'R <-> quaternion transforms are inverses of one another. Broadcasted')

# I'm concerned about quat_from_R() broadcasting properly. I check
testutils.confirm_equal( mrcal.quat_from_R(RR.reshape(2,1,3,3)).shape,
                         (2,1,4),
                         msg = 'quat_from_R() shape')
testutils.confirm_equal( mrcal.quat_from_R(RR.reshape(2,3,3)).shape,
                         (2,4),
                         msg = 'quat_from_R() shape')
testutils.confirm_equal( mrcal.quat_from_R(R).shape,
                         (4,),
                         msg = 'quat_from_R() shape')

# in-place output
R2  = np.zeros(R. shape, dtype=float)
RR2 = np.zeros(RR.shape, dtype=float)
r2  = np.zeros(R. shape[:-2] + (3,), dtype=float)
rr2 = np.zeros(RR.shape[:-2] + (3,), dtype=float)
q2  = np.zeros(R. shape[:-2] + (4,), dtype=float)
qq2 = np.zeros(RR.shape[:-2] + (4,), dtype=float)
q  = mrcal.quat_from_R(R)
qq = mrcal.quat_from_R(RR)
mrcal.quat_from_R(R, out=q2)
testutils.confirm_equal( q2, q,
                         msg = 'quat_from_R() in-place')
mrcal.quat_from_R(RR, out=qq2)
testutils.confirm_equal( qq2, qq,
                         msg = 'quat_from_R() in-place')
mrcal.R_from_quat(q, out=R2)
testutils.confirm_equal( R2,
                         mrcal.R_from_quat(q),
                         msg = 'R_from_quat() in-place')
mrcal.R_from_quat(qq, out=RR2)
testutils.confirm_equal( RR2,
                         mrcal.R_from_quat(qq),
                         msg = 'R_from_quat() in-place')

Rt2   = np.zeros(Rt.  shape, dtype=float)
RtRt2 = np.zeros(RtRt.shape, dtype=float)
rt2   = np.zeros(Rt.  shape[:-2] + (6,), dtype=float)
rtrt2 = np.zeros(RtRt.shape[:-2] + (6,), dtype=float)
qt2   = np.zeros(Rt.  shape[:-2] + (7,), dtype=float)
qtqt2 = np.zeros(RtRt.shape[:-2] + (7,), dtype=float)
qt    = mrcal.qt_from_Rt(Rt)
qtqt  = mrcal.qt_from_Rt(RtRt)
mrcal.qt_from_Rt(Rt, out=qt2)
testutils.confirm_equal( qt2, qt,
                         msg = 'qt_from_Rt() in-place')
mrcal.qt_from_Rt(RtRt, out=qtqt2)
testutils.confirm_equal( qtqt2, qtqt,
                         msg = 'qt_from_Rt() in-place')
mrcal.Rt_from_qt(qt, out=Rt2)
testutils.confirm_equal( Rt2,
                         mrcal.Rt_from_qt(qt),
                         msg = 'Rt_from_qt() in-place')
mrcal.Rt_from_qt(qtqt, out=RtRt2)
testutils.confirm_equal( RtRt2,
                         mrcal.Rt_from_qt(qtqt),
                         msg = 'Rt_from_qt() in-place')


####### small-angle R_from_r()
for th in (1e-12, 1e-11, 1e-10, 1e-9, 1e-8, 1e-7, 1e-6):
    c,s = np.cos(th), np.sin(th)
    rtiny = np.array((0, th, 0))
    Rtiny = np.array((( c, 0, s),
                      ( 0, 1, 0),
                      (-s, 0, c)))

    testutils.confirm_equal( mrcal.R_from_r(rtiny),
                             Rtiny,
                             worstcase = True,
                             eps = 1e-13,
                             msg = f'R_from_r() for tiny rotations: th = {th}')
    testutils.confirm_equal( mrcal.r_from_R(Rtiny),
                             rtiny,
                             worstcase = True,
                             eps = 1e-13,
                             msg = f'r_from_R() for tiny rotations: th = {th}')


###### skew_symmetric()
a = np.array((1.,5.,7.))
b = np.array((3.,-.1,-10.))

A = mrcal.skew_symmetric(a)
testutils.confirm_equal( nps.inner(A,b),
                         np.cross(a,b),
                         eps = 1e-13,
                         msg = f'skew_symmetric()')


testutils.finish()



#!/usr/bin/python3

r'''Tests the CHOLMOD_factorization python class'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils

from scipy.sparse import csr_matrix

indptr  = np.array([0, 2, 3, 6, 8])
indices = np.array([0, 2, 2, 0, 1, 2, 1, 2])
data    = np.array([1, 2, 3, 4, 5, 6, 7, 8], dtype=float)

Jsparse = csr_matrix((data, indices, indptr))
Jdense  = Jsparse.toarray()
Jdense_ref = \
    np.array(((1, 0, 2),
              (0, 0, 3),
              (4, 5, 6),
              (0, 7, 8)), dtype=float)

testutils.confirm_equal(Jdense,
                        Jdense_ref,
                        relative  = True,
                        worstcase = True,
                        eps       = 1e-6,
                        msg       = "csr_matrix representation works as expected")

bt  = np.array(((1., 5., 3.), (2., -2., -8)))

F  = mrcal.CHOLMOD_factorization(Jsparse)
xt = F.solve_xt_JtJ_bt(bt)

JtJ    = nps.matmult(nps.transpose(Jdense), Jdense)
xt_ref = nps.transpose(np.linalg.solve(JtJ, nps.transpose(bt)))

testutils.confirm_equal(xt, xt_ref,
                        relative  = True,
                        worstcase = True,
                        eps       = 1e-6,
                        msg       = "solve_xt_JtJ_bt produces the correct result")

testutils.finish()



r'''Translation of fpbpl and fpbisp from Fortran to Python

The scipy source includes FITPACK, which implements b-spline interpolation. I
converted its 2D surface interpolation routine (fpbisp) to C (via f2c), and then
semi-manually to Python. I can then feed it sympy symbols, and get out
analytical expressions, which are surprisingly difficult to find in a nice
usable form on the internet.

The original fpbisp implementation is at

  scipy/interpolate/fitpack/fpbisp.f

The analysis that uses this conversion lives in bsplines.py

'''

import sympy

def fpbspl_(t, k, x, l):

    h__ = [None] * 100
    hh  = [None] * 100

    h__[-1 + 1] = sympy.Integer(1)
    i__1 = k
    for j in range(1,i__1+1):
        i__2 = j
        for i__ in range(1,i__2+1):
            hh[i__ - 1] = h__[-1 + i__]

        h__[-1 + 1] = 0
        i__2 = j
        for i__ in range(1,i__2+1):
            li = l + i__
            lj = li - j
            f = hh[i__ - 1] / (t[li] - t[lj])
            h__[-1 + i__] += f * (t[li] - x)
            h__[-1 + i__ + 1] = f * (x - t[lj])

    return h__


# argx is between tx[lx] and tx[lx+1]. Same with y
def fpbisp_(tx, ty, k, c, argx, lx, argy, ly):

    wx = [None] * 100
    wy = [None] * 100

    h__ = fpbspl_(tx, k, argx, lx)
    for j in range(1, (k+1)+1):
        wx[1 + j] = h__[j - 1]

    h__ = fpbspl_(ty, k, argy, ly)
    for j  in range(1,(k+1)+1):
        wy[1 + j]   = h__[j - 1]


    for i1 in range(1,(k+1)+1):
        h__[i1 - 1] = wx[1 + i1]

    sp = 0
    for i1 in range(1,(k+1)+1):
        for j1 in range(1,(k+1)+1):
            sp += c[j1-1,i1-1] * h__[i1 - 1] * wy[1 + j1]

    return sp
